#define _GNU_SOURCE
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <err.h>
#include <stdint.h>
#include <inttypes.h>
#include <sys/syscall.h>
#include <sys/timerfd.h>
#include <sys/mman.h>
#include <sys/epoll.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

struct cachestat_range {
  uint64_t off, len;
};
struct cachestat {
  uint64_t nr_cache, nr_dirty, nr_writeback, nr_evicted, nr_recently_evicted;
};

static int fd_main, fd_otherfile;

static void touch_page(int fd, long i) {
  SYSCHK(pwrite(fd, "A", 1, i*0x1000));
}
static void flush_lruvec() {
  SYSCHK(munmap(SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0)), 0x1000));
}
static void zap_page(int fd, int i) {
  SYSCHK(fallocate(fd, FALLOC_FL_PUNCH_HOLE|FALLOC_FL_KEEP_SIZE, i*0x1000, 0x1000));
}

static void epoll_add(int epfd, int fd) {
  struct epoll_event ev = { .events = EPOLLIN };
  SYSCHK(epoll_ctl(epfd, EPOLL_CTL_ADD, fd, &ev));
}

struct timespec get_mono_time(void) {
  struct timespec ts;
  SYSCHK(clock_gettime(CLOCK_MONOTONIC, &ts));
  return ts;
}

static void *thread_fn(void *dummy) {
  while (1) {
    touch_page(fd_main, 1);
    flush_lruvec();
    zap_page(fd_main, 1);

    touch_page(fd_otherfile, 0x12345);
    zap_page(fd_otherfile, 0x12345);
  }
}

int main(void) {
  fd_main = SYSCHK(memfd_create("", 0));
  SYSCHK(ftruncate(fd_main, 0x2000));
  fd_otherfile = SYSCHK(memfd_create("", 0));
  SYSCHK(ftruncate(fd_otherfile, 0x20000 * 0x1000UL));
  touch_page(fd_main, 0);

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "pthread_create");

  while (1) {
    struct cachestat_range range = { .off = 0, .len = 0x2000};
    struct cachestat cstat = {};
    if (syscall(451/*__NR_cachestat*/, fd_main, &range, &cstat, 0))
      err(1, "cachestat");
    if (cstat.nr_cache > 2)
      printf("cstat.nr_cache = 0x%016"PRIx64" (-0x%016"PRIx64")\n", cstat.nr_cache, -cstat.nr_cache);
  }
}
